%   MATLAB script: 
%   3-D reconstruction for rotational symmetry
%   June '08, CVPR 2008 Tutorial on Computational Symmetry
%
%   Author: Allen Y. Yang - yang@eecs.berkeley.edu
%
%   Reference: An Invitation to 3-D Vision, MaSKS, Springer.
%
%   This package is free to use without restrictions. No support is
%   guaranteed.

close all
clear
clc

addpath helper_functions

% Step 1: set up the scene
Rxy = [cos(2*pi/5) -sin(2*pi/5) 0 ; sin(2*pi/5) cos(2*pi/5) 0; 0 0 1];
X(1:3,1) = [0;1;0];
for index=2:5
    X(1:3,index) = Rxy*X(1:3,index-1);
end

R0 = [cos(pi/10) 0 -sin(pi/10); 0 1 0; sin(pi/10) 0 cos(pi/10)];
T0 = [2;3;1];

% Step 2: Perspective projection
figure(1); 
subplot(1,2,1); hold on
set(gcf,'color','white');
colorArray = ['k','r','b','y','g'];
for index=1:5
    x1(1:3,index) = R0*X(:,index)+T0;
    x1(1:3,index) = x1(1:3,index)/x1(3,index);
    plot(x1(1,index),x1(2,index),[colorArray(index) '*'],'MarkerSize',10);
end

% Step 3: Generate a hidden view
x2(:,1) = x1(:,2); x2(:,2) = x1(:,3); x2(:,3) = x1(:,4); 
x2(:,4)=x1(:,5); x2(:,5) = x1(:,1);
subplot(1,2,2); hold on
for index=1:5
    plot(x2(1,index),x2(2,index),[colorArray(index) '*'],'MarkerSize',10);
end

% Step 4: Four-point algorithm
sampleCount = 5;
A = zeros(3*sampleCount,9);
for sampleIndex=1:sampleCount
    A((sampleIndex-1)*3+1:sampleIndex*3,:) = kron(x1(:,sampleIndex),hat(x2(:,sampleIndex)))';
end
[U, S, V] = svd(A);
Hstacked = V(:,end);
H = reshape(Hstacked,[3 3]);

[U, S, V] = svd(H);
H = H/S(2,2);

signArray = zeros(1,sampleCount);
for sampleIndex=1:sampleCount
    signArray(sampleIndex) = sign(x2(:,sampleIndex).'*H*x1(:,sampleIndex));
end
if any(signArray~=signArray(1))
    warning('Tested pairs do not have uniform depth sign!');
end
if sum(signArray)<0
    H = -H;
end

H

% Step 5: Compute R, t, N
solutions= cell(1,4);
[U, S, V] = svd(H.'*H);
v1 = V(:,1)/norm(V(:,1)); v2 = V(:,2)/norm(V(:,2)); v3 = V(:,3)/norm(V(:,3));
u1 = (sqrt(1-S(3,3))*v1+sqrt(S(1,1)-1)*v3)/sqrt(S(1,1)-S(3,3));
u2 = (sqrt(1-S(3,3))*v1-sqrt(S(1,1)-1)*v3)/sqrt(S(1,1)-S(3,3));

solutions{1}.R = [H*v2, H*u1, hat(H*v2)*H*u1]*[v2, u1, hat(v2)*u1].';
solutions{1}.N = hat(v2)*u1;
solutions{1}.t = (H-solutions{1}.R)*solutions{1}.N;
solutions{1}.t = solutions{1}.t/norm(solutions{1}.t);
solutions{2}.R = solutions{1}.R;
solutions{2}.N = -solutions{1}.N;
solutions{2}.t = -solutions{1}.t;
solutions{3}.R = [H*v2, H*u2, hat(H*v2)*H*u2]*[v2, u2, hat(v2)*u2].';
solutions{3}.N = hat(v2)*u2;
solutions{3}.t = (H-solutions{3}.R)*solutions{3}.N;
solutions{3}.t = solutions{3}.t/norm(solutions{3}.t);
solutions{4}.R = solutions{3}.R;
solutions{4}.N = -solutions{3}.N;
solutions{4}.t = -solutions{3}.t;

% Test positive depth constraint and symmetry constraint
figure(2);
subplotIndex=0;
for solutionIndex=1:4
    if solutions{solutionIndex}.N(3)>=0
        % positive depth passed
        R = solutions{solutionIndex}.R;
        t = solutions{solutionIndex}.t;
        
        % compute the missing depth for all points
        M = zeros(3*5,6);
        M(1:3,1) = hat(x2(:,1))*R*x1(:,1);
        M(1:3,6) = hat(x2(:,1))*t;
        M(4:6,2) = hat(x2(:,2))*R*x1(:,2);
        M(4:6,6) = hat(x2(:,2))*t;
        M(7:9,3) = hat(x2(:,3))*R*x1(:,3);
        M(7:9,6) = hat(x2(:,3))*t;
        M(10:12,4) = hat(x2(:,4))*R*x1(:,4);
        M(10:12,6) = hat(x2(:,4))*t;
        M(13:15,5) = hat(x2(:,5))*R*x1(:,5);
        M(13:15,6) = hat(x2(:,5))*t;
        [U,S,V] = svd(M);
        currentX = zeros(3,5);
        for index=1:5
            currentX(:,index) = abs(V(index,end))*x1(:,index);
        end
        if solutionIndex==1
            % I know that this is the true solution
            estimatedX = currentX;
        end
        
        % plot the reconstructed image
        subplotIndex=subplotIndex+1;
        subplot(1,2,subplotIndex); hold on
        plot3(currentX(1,:), currentX(2,:), currentX(3,:),'ko','MarkerSize',5);
        axis equal
    end
end

disp('Notice that only solution #1 gives back a symmetric pentagon after reprojection!');

R=solutions{1}.R
t=solutions{1}.t
N=solutions{1}.N

% Step 5: Compute R0, t0
t0 = mean(estimatedX,2)
[V,D] = eig(R);
v1 = -imag(V(:,2));
v1 = v1 /norm(v1);
v2 = real(V(:,2));
v2 = v2/norm(v2);
v3 = V(:,3);
R0 = [v1, v2,v3]